#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["LinkLabel", "EditLabel", "BasicLabel", "MarkupLabel", "DualMarkupLabel"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
from html import escape
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import logging

_LOG = logging.getLogger(".widgets.labels")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.constfunc import has_display, win
from ..utils import get_link_color

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
if has_display():
    HAND_CURSOR = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.HAND2
    )


# -------------------------------------------------------------------------
#
# Module functions
#
# -------------------------------------------------------------------------
def realize_cb(widget):
    widget.get_window().set_cursor(HAND_CURSOR)


# -------------------------------------------------------------------------
#
# LinkLabel class
#
# -------------------------------------------------------------------------
class LinkLabel(Gtk.EventBox):
    def __init__(self, label, func, handle, emph=False, theme="CLASSIC"):
        self.theme = theme
        self.emph = emph

        Gtk.EventBox.__init__(self)

        st_cont = self.get_style_context()
        self.color = get_link_color(st_cont)

        if emph:
            # emphasize a link
            if theme == "CLASSIC":
                format = 'underline="single" weight="heavy" style="italic"'
            elif theme == "WEBPAGE":
                format = 'foreground="' + self.color + '" weight="heavy"'
            else:
                raise AttributeError("invalid theme: '%s'" % theme)
        elif emph is None:
            # emphasize, but not a link
            if theme == "CLASSIC":
                format = 'weight="heavy"'
            elif theme == "WEBPAGE":
                format = 'weight="heavy"'
            else:
                raise AttributeError("invalid theme: '%s'" % theme)
        else:
            # no emphasize, a link
            if theme == "CLASSIC":
                format = 'underline="single"'
            elif theme == "WEBPAGE":
                format = 'foreground="' + self.color + '"'
            else:
                raise AttributeError("invalid theme: '%s'" % theme)

        self.orig_text = escape(label[0])
        self.gender = label[1]
        self.decoration = format
        text = "<span %s>%s</span>" % (self.decoration, self.orig_text)

        if func:
            msg = _(
                "Click to make this person active\n"
                "Right click to display the edit menu\n"
                "Click Edit icon (enable in configuration dialog) to edit"
            )

            self.set_tooltip_text(msg)

        self.label = Gtk.Label(label=text)
        self.label.set_use_markup(True)
        self.label.set_halign(Gtk.Align.START)

        hbox = Gtk.Box()
        hbox.pack_start(self.label, False, False, 0)
        if label[1]:
            hbox.pack_start(Gtk.Label(label=label[1]), False, False, 0)
            hbox.set_spacing(4)
        self.add(hbox)

        if func:
            self.connect("button-press-event", func, handle)
            self.connect("enter-notify-event", self.enter_text, handle)
            self.connect("leave-notify-event", self.leave_text, handle)
            self.connect("realize", realize_cb)

    def set_padding(self, x, y):
        self.label.set_margin_start(x)
        self.label.set_margin_end(x)
        self.label.set_margin_top(x)
        self.label.set_margin_bottom(x)

    def enter_text(self, obj, event, handle):
        if self.emph:
            # emphasize a link
            if self.theme == "CLASSIC":
                format = (
                    'foreground="' + self.color + '" underline="single" '
                    'weight="heavy" style="italic"'
                )
            elif self.theme == "WEBPAGE":
                format = (
                    'underline="single" foreground="' + self.color + '" '
                    'weight="heavy"'
                )
            else:
                raise AttributeError("invalid theme: '%s'" % self.theme)
        elif self.emph is None:
            # no link, no change on enter_text
            if self.theme == "CLASSIC":
                format = 'weight="heavy"'
            elif self.theme == "WEBPAGE":
                format = 'weight="heavy"'
            else:
                raise AttributeError("invalid theme: '%s'" % self.theme)
        else:
            # no emphasize, a link
            if self.theme == "CLASSIC":
                format = 'foreground="' + self.color + '" underline="single"'
            elif self.theme == "WEBPAGE":
                format = 'underline="single" foreground="' + self.color + '"'
            else:
                raise AttributeError("invalid theme: '%s'" % self.theme)

        text = "<span %s>%s</span>" % (format, self.orig_text)
        self.label.set_text(text)
        self.label.set_use_markup(True)

    def leave_text(self, obj, event, handle):
        text = "<span %s>%s</span>" % (self.decoration, self.orig_text)
        self.label.set_text(text)
        self.label.set_use_markup(True)


# -------------------------------------------------------------------------
#
# EditLabel class
#
# -------------------------------------------------------------------------
class EditLabel(Gtk.Box):
    def __init__(self, text):
        Gtk.Box.__init__(self)
        label = BasicLabel(text)
        self.pack_start(label, False, True, 0)
        self.pack_start(
            Gtk.Image.new_from_icon_name("gtk-edit", Gtk.IconSize.MENU), False
        )
        self.set_spacing(4)
        self.show_all()


# -------------------------------------------------------------------------
#
# BasicLabel class
#
# -------------------------------------------------------------------------
class BasicLabel(Gtk.Label):
    def __init__(self, text, ellipsize=Pango.EllipsizeMode.NONE):
        Gtk.Label.__init__(self, label=text)
        self.set_halign(Gtk.Align.START)
        self.set_ellipsize(ellipsize)
        self.show()


# -------------------------------------------------------------------------
#
# MarkupLabel class
#
# -------------------------------------------------------------------------
class MarkupLabel(Gtk.Label):
    def __init__(self, text, halign=Gtk.Align.START):
        Gtk.Label.__init__(self, label=text)
        self.set_halign(halign)
        self.set_use_markup(True)
        self.show_all()


# -------------------------------------------------------------------------
#
# DualMarkupLabel class
#
# -------------------------------------------------------------------------
class DualMarkupLabel(Gtk.Box):
    def __init__(self, text, alt, halign=Gtk.Align.START):
        Gtk.Box.__init__(self)
        label = Gtk.Label(label=text)
        label.set_halign(halign)
        label.set_use_markup(True)

        self.pack_start(label, False, False, 0)
        b = Gtk.Label(label=alt)
        self.pack_start(b, False, False, 4)
        self.show()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# $Id: categorysidebar.py 20634 2012-11-07 17:53:14Z bmcage $

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.config import config
from gramps.gui.basesidebar import BaseSidebar
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# DropdownSidebar class
#
# -------------------------------------------------------------------------
class DropdownSidebar(BaseSidebar):
    """
    A sidebar displaying toggle buttons and buttons with drop-down menus that
    allows the user to change the current view.
    """

    def __init__(self, dbstate, uistate, categories, views):
        self.viewmanager = uistate.viewmanager
        self.views = views

        self.buttons = []
        self.button_handlers = []

        self.window = Gtk.ScrolledWindow()
        grid = Gtk.Grid()
        self.window.add(grid)
        self.window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.window.show()

        use_text = config.get("interface.sidebar-text")
        for cat_num, cat_name, cat_icon in categories:
            self.__make_category(grid, use_text, cat_num, cat_name, cat_icon)

        grid.show_all()

    def get_top(self):
        """
        Return the top container widget for the GUI.
        """
        return self.window

    def view_changed(self, cat_num, view_num):
        """
        Called when the active view is changed.
        """
        # Set new button as selected
        self.__handlers_block()
        for index, button in enumerate(self.buttons):
            if index == cat_num:
                button.set_active(True)
            else:
                button.set_active(False)
        self.__handlers_unblock()

    def __handlers_block(self):
        """
        Block signals to the buttons to prevent spurious events.
        """
        for idx, button in enumerate(self.buttons):
            button.handler_block(self.button_handlers[idx])

    def __handlers_unblock(self):
        """
        Unblock signals to the buttons.
        """
        for idx, button in enumerate(self.buttons):
            button.handler_unblock(self.button_handlers[idx])

    def cb_view_clicked(self, radioaction, current, cat_num):
        """
        Called when a button causes a view change.
        """
        view_num = radioaction.get_current_value()
        self.viewmanager.goto_page(cat_num, view_num)

    def __category_clicked(self, button, cat_num):
        """
        Called when a category button is clicked.
        """
        # Make the button active.  If it was already active the category will
        # not change.
        button.set_active(True)
        self.viewmanager.goto_page(cat_num, None)

    def __view_clicked(self, button, cat_num):
        """
        Called when a view drop-down arrow is clicked.
        """
        self.menu = Gtk.Menu()
        self.menu.set_reserve_toggle_size(False)
        for item in self.views[cat_num]:
            menuitem = Gtk.MenuItem()
            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            label = Gtk.Label(label=item[1])
            image = Gtk.Image.new_from_icon_name(item[2], Gtk.IconSize.MENU)
            hbox.pack_start(image, False, False, 3)
            hbox.pack_start(label, False, False, 3)
            hbox.show_all()
            menuitem.add(hbox)
            menuitem.connect("activate", self.cb_menu_clicked, cat_num, item[0])
            menuitem.show()
            self.menu.append(menuitem)
        self.menu.popup_at_widget(
            button, Gdk.Gravity.SOUTH_WEST, Gdk.Gravity.NORTH_WEST, None
        )

    def cb_menu_clicked(self, menuitem, cat_num, view_num):
        """
        Called when a view is selected from a drop-down menu.
        """
        self.viewmanager.goto_page(cat_num, view_num)

    def __make_category(self, grid, use_text, cat_num, cat_name, cat_icon):
        """
        Create a row in the sidebar for a category.
        """
        # create the button
        button = Gtk.ToggleButton()
        button.set_relief(Gtk.ReliefStyle.NONE)
        self.buttons.append(button)

        # create the drop-down button to display views
        if len(self.views[cat_num]) > 1:
            dropdown = Gtk.Button()
            dropdown.set_relief(Gtk.ReliefStyle.NONE)
            arrow = Gtk.Arrow(
                arrow_type=Gtk.ArrowType.DOWN, shadow_type=Gtk.ShadowType.NONE
            )
            dropdown.add(arrow)
            dropdown.connect("clicked", self.__view_clicked, cat_num)
            dropdown.set_tooltip_text(_("Click to select a view"))
            grid.attach(dropdown, 1, cat_num, 1, 1)

        # add the tooltip
        button.set_tooltip_text(cat_name)

        # connect the signal, along with the cat_num as user data
        handler_id = button.connect("clicked", self.__category_clicked, cat_num)
        self.button_handlers.append(handler_id)
        button.show()

        # add the image. If we are using text, use the BUTTON (larger) size.
        # otherwise, use the smaller size
        hbox = Gtk.Box()
        hbox.show()
        image = Gtk.Image()
        if use_text:
            image.set_from_icon_name(cat_icon, Gtk.IconSize.BUTTON)
        else:
            image.set_from_icon_name(cat_icon, Gtk.IconSize.DND)
        image.show()
        hbox.pack_start(image, False, False, 0)
        hbox.set_spacing(4)

        # add text if requested
        if use_text:
            label = Gtk.Label(label=cat_name)
            label.show()
            hbox.pack_start(label, False, True, 0)

        button.add(hbox)

        # Enable view switching during DnD
        button.drag_dest_set(0, [], 0)
        button.connect("drag_motion", self.cb_switch_page_on_dnd, cat_num)

        grid.attach(button, 0, cat_num, 1, 1)

    def cb_switch_page_on_dnd(self, widget, context, xpos, ypos, time, page_no):
        """
        Switches the page based on drag and drop.
        """
        self.__handlers_block()
        if self.viewmanager.notebook.get_current_page() != page_no:
            self.viewmanager.notebook.set_current_page(page_no)
        self.__handlers_unblock()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010,2017  Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Family object for Gramps.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
from warnings import warn
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .primaryobj import PrimaryObject
from .citationbase import CitationBase
from .notebase import NoteBase
from .mediabase import MediaBase
from .attrbase import AttributeBase
from .eventref import EventRef
from .ldsordbase import LdsOrdBase
from .tagbase import TagBase
from .childref import ChildRef
from .familyreltype import FamilyRelType
from .const import IDENTICAL, EQUAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

LOG = logging.getLogger(".citation")


# -------------------------------------------------------------------------
#
# Family class
#
# -------------------------------------------------------------------------
class Family(
    CitationBase, NoteBase, MediaBase, AttributeBase, LdsOrdBase, PrimaryObject
):
    """
    The Family record is the Gramps in-memory representation of the
    relationships between people. It contains all the information
    related to the relationship.

    Family objects are usually created in one of two ways.

    1. Creating a new Family object, which is then initialized and
       added to the database.
    2. Retrieving an object from the database using the records
       handle.

    Once a Family object has been modified, it must be committed
    to the database using the database object's commit_family function,
    or the changes will be lost.
    """

    def __init__(self):
        """
        Create a new Family instance.

        After initialization, most data items have empty or null values,
        including the database handle.
        """
        PrimaryObject.__init__(self)
        CitationBase.__init__(self)
        NoteBase.__init__(self)
        MediaBase.__init__(self)
        AttributeBase.__init__(self)
        LdsOrdBase.__init__(self)
        self.father_handle = None
        self.mother_handle = None
        self.child_ref_list = []
        self.type = FamilyRelType()
        self.event_ref_list = []
        self.complete = 0

    def serialize(self):
        """
        Convert the data held in the event to a Python tuple that
        represents all the data elements.

        This method is used to convert the object into a form that can easily
        be saved to a database.

        These elements may be primitive Python types (string, integers),
        complex Python types (lists or tuples, or Python objects. If the
        target database cannot handle complex types (such as objects or
        lists), the database is responsible for converting the data into
        a form that it can use.

        :returns: Returns a python tuple containing the data that should
                  be considered persistent.
        :rtype: tuple
        """
        return (
            self.handle,
            self.gramps_id,
            self.father_handle,
            self.mother_handle,
            [cr.serialize() for cr in self.child_ref_list],
            self.type.serialize(),
            [er.serialize() for er in self.event_ref_list],
            MediaBase.serialize(self),
            AttributeBase.serialize(self),
            LdsOrdBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            self.change,
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .mediaref import MediaRef
        from .ldsord import LdsOrd
        from .childref import ChildRef
        from .attribute import Attribute

        return {
            "type": "object",
            "title": _("Family"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {"type": "string", "maxLength": 50, "title": _("Handle")},
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "father_handle": {
                    "type": ["string", "null"],
                    "maxLength": 50,
                    "title": _("Father"),
                },
                "mother_handle": {
                    "type": ["string", "null"],
                    "maxLength": 50,
                    "title": _("Mother"),
                },
                "child_ref_list": {
                    "type": "array",
                    "items": ChildRef.get_schema(),
                    "title": _("Children"),
                },
                "type": FamilyRelType.get_schema(),
                "event_ref_list": {
                    "type": "array",
                    "items": EventRef.get_schema(),
                    "title": _("Events"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "attribute_list": {
                    "type": "array",
                    "items": Attribute.get_schema(),
                    "title": _("Attributes"),
                },
                "lds_ord_list": {
                    "type": "array",
                    "items": LdsOrd.get_schema(),
                    "title": _("LDS ordinances"),
                },
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Family structure.
        """
        (
            self.handle,
            self.gramps_id,
            self.father_handle,
            self.mother_handle,
            child_ref_list,
            the_type,
            event_ref_list,
            media_list,
            attribute_list,
            lds_seal_list,
            citation_list,
            note_list,
            self.change,
            tag_list,
            self.private,
        ) = data

        self.type = FamilyRelType()
        self.type.unserialize(the_type)
        self.event_ref_list = [EventRef().unserialize(er) for er in event_ref_list]
        self.child_ref_list = [ChildRef().unserialize(cr) for cr in child_ref_list]
        MediaBase.unserialize(self, media_list)
        AttributeBase.unserialize(self, attribute_list)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        LdsOrdBase.unserialize(self, lds_seal_list)
        TagBase.unserialize(self, tag_list)
        return self

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                  this object type.
        :rtype: bool
        """
        if classname == "Event":
            return handle in [ref.ref for ref in self.event_ref_list]
        elif classname == "Person":
            return handle in (
                [ref.ref for ref in self.child_ref_list]
                + [self.father_handle, self.mother_handle]
            )
        elif classname == "Place":
            return handle in [x.place for x in self.lds_ord_list]
        return False

    def _remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """
        if classname == "Event":
            new_list = [
                ref for ref in self.event_ref_list if ref.ref not in handle_list
            ]
            self.event_ref_list = new_list
        elif classname == "Person":
            new_list = [
                ref for ref in self.child_ref_list if ref.ref not in handle_list
            ]
            self.child_ref_list = new_list
            if self.father_handle in handle_list:
                self.father_handle = None
            if self.mother_handle in handle_list:
                self.mother_handle = None
        elif classname == "Place":
            for lds_ord in self.lds_ord_list:
                if lds_ord.place in handle_list:
                    lds_ord.place = None

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Event":
            refs_list = [ref.ref for ref in self.event_ref_list]
            new_ref = None
            if new_handle in refs_list:
                new_ref = self.event_ref_list[refs_list.index(new_handle)]
            n_replace = refs_list.count(old_handle)
            for ix_replace in range(n_replace):
                idx = refs_list.index(old_handle)
                self.event_ref_list[idx].ref = new_handle
                refs_list[idx] = new_handle
                if new_ref:
                    evt_ref = self.event_ref_list[idx]
                    equi = new_ref.is_equivalent(evt_ref)
                    if equi != DIFFERENT:
                        if equi == EQUAL:
                            new_ref.merge(evt_ref)
                        self.event_ref_list.pop(idx)
                        refs_list.pop(idx)
        elif classname == "Person":
            refs_list = [ref.ref for ref in self.child_ref_list]
            new_ref = None
            if new_handle in refs_list:
                new_ref = self.child_ref_list[refs_list.index(new_handle)]
            n_replace = refs_list.count(old_handle)
            for ix_replace in range(n_replace):
                idx = refs_list.index(old_handle)
                self.child_ref_list[idx].ref = new_handle
                refs_list[idx] = new_handle
                if new_ref:
                    child_ref = self.child_ref_list[idx]
                    equi = new_ref.is_equivalent(child_ref)
                    if equi != DIFFERENT:
                        if equi == EQUAL:
                            new_ref.merge(child_ref)
                        self.child_ref_list.pop(idx)
                        refs_list.pop(idx)
            if self.father_handle == old_handle:
                self.father_handle = new_handle
            if self.mother_handle == old_handle:
                self.mother_handle = new_handle
        elif classname == "Place":
            for lds_ord in self.lds_ord_list:
                if lds_ord.place == old_handle:
                    lds_ord.place = new_handle

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        add_list = [_f for _f in self.lds_ord_list if _f]
        return self.media_list + self.attribute_list + add_list

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        check_list = (
            self.media_list
            + self.attribute_list
            + self.lds_ord_list
            + self.child_ref_list
            + self.event_ref_list
        )
        return check_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        check_list = (
            self.media_list
            + self.attribute_list
            + self.lds_ord_list
            + self.child_ref_list
            + self.event_ref_list
        )
        return check_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )
        ret += [
            ("Person", handle)
            for handle in (
                [ref.ref for ref in self.child_ref_list]
                + [self.father_handle, self.mother_handle]
            )
            if handle
        ]
        ret += self.get_referenced_tag_handles()
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through their
        children, reference primary objects..

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return (
            self.media_list
            + self.attribute_list
            + self.lds_ord_list
            + self.child_ref_list
            + self.event_ref_list
        )

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this family.

        Lost: handle, id, relation, father, mother of acquisition.

        :param acquisition: The family to merge with the present family.
        :type acquisition: Family
        """
        if self.type != acquisition.type and self.type == FamilyRelType.UNKNOWN:
            self.set_relationship(acquisition.get_relationship())
        self._merge_privacy(acquisition)
        self._merge_event_ref_list(acquisition)
        self._merge_lds_ord_list(acquisition)
        self._merge_media_list(acquisition)
        self._merge_child_ref_list(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_tag_list(acquisition)

    def set_relationship(self, relationship_type):
        """
        Set the relationship type between the people identified as the
        father and mother in the relationship.

        The type is a tuple whose first item is an integer constant and whose
        second item is the string. The valid values are:

        =========================  ============================================
        Type                       Description
        =========================  ============================================
        FamilyRelType.MARRIED      indicates a legally recognized married
                                   relationship between two individuals. This
                                   may be either an opposite or a same sex
                                   relationship.
        FamilyRelType.UNMARRIED    indicates a relationship between two
                                   individuals that is not a legally recognized
                                   relationship.
        FamilyRelType.CIVIL_UNION  indicates a legally recongnized, non-married
                                   relationship between two individuals of the
                                   same sex.
        FamilyRelType.UNKNOWN      indicates that the type of relationship
                                   between the two individuals is not know.
        FamilyRelType.CUSTOM       indicates that the type of relationship
                                   between the two individuals does not match
                                   any of the other types.
        =========================  ============================================

        :param relationship_type: (int,str) tuple of the relationship type
               between the father and mother of the relationship.
        :type relationship_type: tuple
        """
        self.type.set(relationship_type)

    def get_relationship(self):
        """
        Return the relationship type between the people identified as the
        father and mother in the relationship.
        """
        return self.type

    def set_father_handle(self, person_handle):
        """
        Set the database handle for :class:`~.person.Person` that corresponds
        to male of the relationship.

        For a same sex relationship, this can represent either of people
        involved in the relationship.

        :param person_handle: :class:`~.person.Person` database handle
        :type person_handle: str
        """
        self.father_handle = person_handle

    def get_father_handle(self):
        """
        Return the database handle of the :class:`~.person.Person` identified
        as the father of the Family.

        :returns: :class:`~.person.Person` database handle
        :rtype: str
        """
        return self.father_handle

    def set_mother_handle(self, person_handle):
        """
        Set the database handle for :class:`~.person.Person` that corresponds
        to male of the relationship.

        For a same sex relationship, this can represent either of people
        involved in the relationship.

        :param person_handle: :class:`~.person.Person` database handle
        :type person_handle: str
        """
        self.mother_handle = person_handle

    def get_mother_handle(self):
        """
        Return the database handle of the :class:`~.person.Person` identified
        as the mother of the Family.

        :returns: :class:`~.person.Person` database handle
        :rtype: str
        """
        return self.mother_handle

    def add_child_ref(self, child_ref):
        """
        Add the database handle for :class:`~.person.Person` to the Family's
        list of children.

        :param child_ref: Child Reference instance
        :type  child_ref: ChildRef
        """
        if not isinstance(child_ref, ChildRef):
            raise ValueError("expecting ChildRef instance")
        self.child_ref_list.append(child_ref)

    def remove_child_ref(self, child_ref):
        """
        Remove the database handle for :class:`~.person.Person` to the Family's
        list of children if the :class:`~.person.Person` is already in the list.

        :param child_ref: Child Reference instance
        :type child_ref: ChildRef
        :returns: True if the handle was removed, False if it was not
                  in the list.
        :rtype: bool
        """
        if not isinstance(child_ref, ChildRef):
            raise ValueError("expecting ChildRef instance")
        new_list = [ref for ref in self.child_ref_list if ref.ref != child_ref.ref]
        self.child_ref_list = new_list

    def remove_child_handle(self, child_handle):
        """
        Remove the database handle for :class:`~.person.Person` to the Family's
        list of children if the :class:`~.person.Person` is already in the list.

        :param child_handle: :class:`~.person.Person` database handle
        :type  child_handle: str
        :returns: True if the handle was removed, False if it was not
                  in the list.
        :rtype: bool
        """
        new_list = [ref for ref in self.child_ref_list if ref.ref != child_handle]
        self.child_ref_list = new_list

    def get_child_ref_list(self):
        """
        Return the list of :class:`~.childref.ChildRef` handles identifying the
        children of the Family.

        :returns: Returns the list of :class:`~.childref.ChildRef` handles
                  associated with the Family.
        :rtype: list
        """
        return self.child_ref_list

    def set_child_ref_list(self, child_ref_list):
        """
        Assign the passed list to the Family's list children.

        :param child_ref_list: List of Child Reference instances to be
                               associated as the Family's list of children.
        :type child_ref_list: list of :class:`~.childref.ChildRef` instances
        """
        self.child_ref_list = child_ref_list

    def _merge_child_ref_list(self, acquisition):
        """
        Merge the list of child references from acquisition with our own.

        :param acquisition: the childref list of this family will be merged
                            with the current childref list.
        :type acquisition: Family
        """
        childref_list = self.child_ref_list[:]
        for addendum in acquisition.get_child_ref_list():
            for childref in childref_list:
                equi = childref.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    childref.merge(addendum)
                    break
            else:
                self.child_ref_list.append(addendum)

    def add_event_ref(self, event_ref):
        """
        Add the :class:`~.eventref.EventRef` to the Family instance's
        :class:`~.eventref.EventRef` list.

        This is accomplished by assigning the :class:`~.eventref.EventRef` for
        the valid :class:`~.event.Event` in the current database.

        :param event_ref: the :class:`~.eventref.EventRef` to be added to the
                          Person's :class:`~.eventref.EventRef` list.
        :type event_ref: EventRef
        """
        if event_ref and not isinstance(event_ref, EventRef):
            raise ValueError("Expecting EventRef instance")
        self.event_ref_list.append(event_ref)

    def get_event_list(self):
        warn("Use get_event_ref_list instead of get_event_list", DeprecationWarning, 2)
        # Wrapper for old API
        # remove when transitition done.
        event_handle_list = []
        for event_ref in self.get_event_ref_list():
            event_handle_list.append(event_ref.get_reference_handle())
        return event_handle_list

    def get_event_ref_list(self):
        """
        Return the list of :class:`~.eventref.EventRef` objects associated with
        :class:`~.event.Event` instances.

        :returns: Returns the list of :class:`~.eventref.EventRef` objects
                  associated with the Family instance.
        :rtype: list
        """
        return self.event_ref_list

    def set_event_ref_list(self, event_ref_list):
        """
        Set the Family instance's :class:`~.eventref.EventRef` list to the
        passed list.

        :param event_ref_list: List of valid :class:`~.eventref.EventRef`
                               objects
        :type event_ref_list: list
        """
        self.event_ref_list = event_ref_list

    def _merge_event_ref_list(self, acquisition):
        """
        Merge the list of event references from acquisition with our own.

        :param acquisition: the event references list of this object will be
                            merged with the current event references list.
        :type acquisition: Person
        """
        eventref_list = self.event_ref_list[:]
        for addendum in acquisition.get_event_ref_list():
            for eventref in eventref_list:
                equi = eventref.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    eventref.merge(addendum)
                    break
            else:
                self.event_ref_list.append(addendum)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Debugging utilities
"""
import cProfile
import pstats
import sys


# -------------------------------------------------------------------------
#
# Debugging utilities
#
# -------------------------------------------------------------------------
def profile(func, *args, **kwargs):
    prf = cProfile.Profile()
    print("Start")
    r = prf.runcall(func, *args, **kwargs)
    print("Finished")
    print("Loading profile")
    stats = pstats.Stats(prf, stream=sys.stdout)
    print("done")
    stats.strip_dirs()
    stats.sort_stats("time", "calls")
    stats.print_stats(100)
    stats.print_callers(100)
    return r


def format_exception(tb_type=None, tb_value=None, tb=None):
    """
    Get the usual traceback information, followed by a listing of all the
    local variables in each frame.
    Based on:
    code.activestate.com/recipes/52215-get-more-information-from-tracebacks
    """
    import sys
    import traceback

    if tb_type is None:
        tb_type = sys.exc_type
    if tb_value is None:
        tb_value = sys.exc_value
    if tb is None:
        tb = sys.exc_info()[2]
    retval = traceback.format_exception(tb_type, tb_value, tb) + ["\n"]
    while tb.tb_next:
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back
    stack.reverse()
    retval.append("Local variables (most recent frame last):\n")
    for frame in stack:
        retval.append(
            ' Frame %s, File "%s", line %s:\n'
            % (frame.f_code.co_name, frame.f_code.co_filename, frame.f_lineno)
        )
        for key, value in frame.f_locals.items():
            if key.startswith("__"):
                continue
            # We have to be careful not to cause a new error in our error
            # handler! Calling str() on an unknown object could cause an
            # error we don't want.
            try:
                line = "  %s = %s\n" % (key, str(value))
            except:
                line = "  %s = %s\n" % (key, "<ERROR PRINTING VALUE>")
            retval.append(line)
    return retval



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Family View.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
import logging

_LOG = logging.getLogger(".plugins.eventview")
# -------------------------------------------------------------------------
#
# GNOME/GTK+ modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Family
from gramps.gui.views.listview import ListView, TEXT, MARKUP, ICON
from gramps.gui.views.treemodels import FamilyModel
from gramps.gui.editors import EditFamily
from gramps.gui.views.bookmarks import FamilyBookmarks
from gramps.gen.errors import WindowActiveError
from gramps.gen.config import config
from gramps.gui.dialog import ErrorDialog
from gramps.gui.filters.sidebar import FamilySidebarFilter
from gramps.gui.merge import MergeFamily
from gramps.gen.plug import CATEGORY_QR_FAMILY
from gramps.gui.ddtargets import DdTargets


# -------------------------------------------------------------------------
#
# FamilyView
#
# -------------------------------------------------------------------------
class FamilyView(ListView):
    """FamilyView class, derived from the ListView"""

    # columns in the model used in view
    COL_ID = 0
    COL_FATHER = 1
    COL_MOTHER = 2
    COL_REL = 3
    COL_MARDATE = 4
    COL_PRIV = 5
    COL_TAGS = 6
    COL_CHAN = 7
    # column definitions
    COLUMNS = [
        (_("ID"), TEXT, None),
        (_("Father"), TEXT, None),
        (_("Mother"), TEXT, None),
        (_("Relationship"), TEXT, None),
        (_("Marriage Date"), MARKUP, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        ("columns.visible", [COL_ID, COL_FATHER, COL_MOTHER, COL_REL, COL_MARDATE]),
        (
            "columns.rank",
            [
                COL_ID,
                COL_FATHER,
                COL_MOTHER,
                COL_REL,
                COL_MARDATE,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
            ],
        ),
        ("columns.size", [75, 200, 200, 100, 100, 40, 100, 100]),
    )

    ADD_MSG = _("Add a new family")
    EDIT_MSG = _("Edit the selected family")
    DEL_MSG = _("Delete the selected family")
    MERGE_MSG = _("Merge the selected families")
    FILTER_TYPE = "Family"
    QR_CATEGORY = CATEGORY_QR_FAMILY

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        signal_map = {
            "family-add": self.row_add,
            "family-update": self.row_update,
            "family-delete": self.row_delete,
            "family-rebuild": self.object_build,
            "event-update": self.related_update,
            "person-update": self.related_update,
        }

        ListView.__init__(
            self,
            _("Families"),
            pdata,
            dbstate,
            uistate,
            FamilyModel,
            signal_map,
            FamilyBookmarks,
            nav_group,
            multiple=True,
            filter_class=FamilySidebarFilter,
        )

        uistate.connect("nameformat-changed", self.build_tree)

        self.additional_uis.append(self.additional_ui)

    def get_config_name(self):
        return __name__

    def navigation_type(self):
        return "Family"

    def get_stock(self):
        return "gramps-family"

    additional_ui = [  # Defines the UI string for UIManager
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Family Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="action">win.MakeFatherActive</attribute>
          <attribute name="label" translatable="yes">"""
        """Make Father Active Person</attribute>
        </item>
        <item>
          <attribute name="action">win.MakeMotherActive</attribute>
          <attribute name="label" translatable="yes">"""
        """Make Mother Active Person</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
      </section>
    </menu>
"""
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def define_actions(self):
        """Add the Forward action group to handle the Forward button."""

        ListView.define_actions(self)

        self.action_list.extend(
            [
                ("MakeFatherActive", self._make_father_active),
                ("MakeMotherActive", self._make_mother_active),
            ]
        )

    def add_bookmark(self, *obj):
        mlist = self.selected_handles()
        if mlist:
            self.bookmarks.add(mlist[0])
        else:
            from gramps.gui.dialog import WarningDialog

            WarningDialog(
                _("Could Not Set a Bookmark"),
                _("A bookmark could not be set because " "no one was selected."),
                parent=self.uistate.window,
            )

    def get_handle_from_gramps_id(self, gid):
        """
        Return the handle of the family having the given Gramps ID.
        """
        obj = self.dbstate.db.get_family_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def add(self, *obj):
        family = Family()
        try:
            EditFamily(self.dbstate, self.uistate, [], family)
        except WindowActiveError:
            pass

    def remove(self, *obj):
        """
        Method called when deleting a family from a family view.
        """
        handles = self.selected_handles()
        ht_list = [("Family", hndl) for hndl in handles]
        self.remove_selected_objects(ht_list)

    def _message1_format(self, family):
        """
        Header format for remove dialogs.
        """
        return _("Delete %s?") % (_("family") + (" [%s]" % family.gramps_id))

    def remove_object_from_handle(
        self, _obj_type, handle, trans, in_use_prompt=False, parent=None
    ):
        """
        deletes a single object from database
        """
        family = self.dbstate.db.get_family_from_handle(handle)
        self.dbstate.db.remove_family_relationships(handle, trans)

    def edit(self, *obj):
        for handle in self.selected_handles():
            family = self.dbstate.db.get_family_from_handle(handle)
            try:
                EditFamily(self.dbstate, self.uistate, [], family)
            except WindowActiveError:
                pass

    def merge(self, *obj):
        """
        Merge the selected families.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge families.")
            msg2 = _(
                "Exactly two families must be selected to perform a merge."
                " A second family can be selected by holding down the "
                "control key while clicking on the desired family."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            MergeFamily(self.dbstate, self.uistate, [], mlist[0], mlist[1])

    def _make_father_active(self, *obj):
        """
        Make the father of the family the active person.
        """
        fhandle = self.first_selected()
        if fhandle:
            family = self.dbstate.db.get_family_from_handle(fhandle)
            if family:
                self.uistate.set_active(family.father_handle, "Person")

    def _make_mother_active(self, *obj):
        """
        Make the mother of the family the active person.
        """
        fhandle = self.first_selected()
        if fhandle:
            family = self.dbstate.db.get_family_from_handle(fhandle)
            if family:
                self.uistate.set_active(family.mother_handle, "Person")

    def drag_info(self):
        """
        Indicate that the drag type is a FAMILY_LINK
        """
        return DdTargets.FAMILY_LINK

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Family"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, family_handle, tag_handle):
        """
        Add the given tag to the given family.
        """
        family = self.dbstate.db.get_family_from_handle(family_handle)
        family.add_tag(tag_handle)
        self.dbstate.db.commit_family(family, transaction)

    def remove_tag(self, transaction, family_handle, tag_handle):
        """
        Remove the given tag from the given family.
        """
        family = self.dbstate.db.get_family_from_handle(family_handle)
        family.remove_tag(tag_handle)
        self.dbstate.db.commit_family(family, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Family Filter",),
            (
                "Family Gallery",
                "Family Events",
                "Family Children",
                "Family Citations",
                "Family Notes",
                "Family Attributes",
                "Family Backlinks",
            ),
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Rebuild Secondary Indexes"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".Rebuild")

# -------------------------------------------------------------------------
#
# gtk modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.plug import tool
from gramps.gui.dialog import OkDialog


# -------------------------------------------------------------------------
#
# runTool
#
# -------------------------------------------------------------------------
class Rebuild(tool.Tool):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate

        tool.Tool.__init__(self, dbstate, options_class, name)

        if self.db.readonly:
            return

        self.db.disable_signals()

        if uistate:
            self.callback = uistate.pulse_progressbar
            uistate.set_busy_cursor(True)
            uistate.progress.show()
            uistate.push_message(dbstate, _("Rebuilding secondary indexes..."))

            self.db.rebuild_secondary(self.callback)

            uistate.set_busy_cursor(False)
            uistate.progress.hide()
            OkDialog(
                _("Secondary indexes rebuilt"),
                _("All secondary indexes have been rebuilt."),
                parent=uistate.window,
            )
        else:
            print("Rebuilding Secondary Indexes...")
            self.db.rebuild_secondary(None)
            print("All secondary indexes have been rebuilt.")

        self.db.enable_signals()
        self.db.request_rebuild()


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class RebuildOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2006-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Person Reference class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .secondaryobj import SecondaryObject
from .privacybase import PrivacyBase
from .citationbase import CitationBase
from .notebase import NoteBase
from .refbase import RefBase
from .const import IDENTICAL, EQUAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Person References for Person/Family
#
# -------------------------------------------------------------------------
class PersonRef(SecondaryObject, PrivacyBase, CitationBase, NoteBase, RefBase):
    """
    Person reference class.

    This class is for keeping information about how the person relates
    to another person from the database, if not through family.
    Examples would be: godparent, friend, etc.
    """

    def __init__(self, source=None):
        PrivacyBase.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        RefBase.__init__(self, source)
        if source:
            self.rel = source.rel
        else:
            self.rel = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            PrivacyBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            RefBase.serialize(self),
            self.rel,
        )

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (privacy, citation_list, note_list, ref, self.rel) = data
        PrivacyBase.unserialize(self, privacy)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        RefBase.unserialize(self, ref)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Person ref"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "citation_list": {
                    "type": "array",
                    "title": _("Citations"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "note_list": {
                    "type": "array",
                    "title": _("Notes"),
                    "items": {"type": "string", "maxLength": 50},
                },
                "ref": {"type": "string", "title": _("Handle"), "maxLength": 50},
                "rel": {"type": "string", "title": _("Association")},
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.rel]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )
        if self.ref:
            ret += [("Person", self.ref)]
        return ret

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects..

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def is_equivalent(self, other):
        """
        Return if this person reference is equivalent, that is agrees in handle
        and relation, to other.

        :param other: The personref to compare this one to.
        :type other: PersonRef
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if (
            self.ref != other.ref
            or self.get_text_data_list() != other.get_text_data_list()
        ):
            return DIFFERENT
        else:
            if self.is_equal(other):
                return IDENTICAL
            else:
                return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this person reference.

        Lost: hlink and relation of acquisition.

        :param acquisition: The personref to merge with the present personref.
        :type acquisition: PersonRef
        """
        self._merge_privacy(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_note_list(acquisition)

    def set_relation(self, rel):
        """Set relation to a person."""
        self.rel = rel

    def get_relation(self):
        """Return the relation to a person."""
        return self.rel



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide a simplified database access interface to the Gramps database.
"""
from ..plug.docgen import (
    StyleSheet,
    ParagraphStyle,
    TableStyle,
    TableCellStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_LEFT,
)


class SimpleDoc:
    """
    Provide a simplified database access interface to the Gramps database.
    """

    def __init__(self, doc):
        """
        Initialize the class with the real document
        """
        self.doc = doc

    def __write(self, format, text):
        """
        Writes a paragraph using the specified format to the BaseDoc
        """
        self.doc.start_paragraph(format)
        self.doc.write_text(text)
        self.doc.end_paragraph()

    def title(self, text):
        """
        Writes the Title using the Title paragraph
        """
        self.__write("Title", text)

    def header1(self, text):
        """
        Writes the first level header using the Header1 paragraph
        """
        self.__write("Header1", text)

    def header2(self, text):
        """
        Writes the second level header using the Header2 paragraph
        """
        self.__write("Header2", text)

    def header3(self, text):
        """
        Writes the third level header using the Header3 paragraph
        """
        self.__write("Header3", text)

    def paragraph(self, text):
        """
        Writes a paragraph using the Normal format
        """
        self.__write("Normal", text)


def make_basic_stylesheet(**kwargs):
    """
    Create the basic style sheet for the SimpleDoc class.

    kwargs - a dictionary of the form:
           item={method: value, ...}, ...

    Example:

    make_basic_stylesheet(Table={"set_width": 90})

    """
    sheet = StyleSheet()

    pstyle = ParagraphStyle()
    fstyle = pstyle.get_font()
    fstyle.set_type_face(FONT_SANS_SERIF)
    fstyle.set_size(14)
    fstyle.set_bold(True)
    pstyle.set_font(fstyle)
    pstyle.set_alignment(PARA_ALIGN_LEFT)
    # Handle args:
    if "Title" in kwargs:
        for method in kwargs["Title"]:
            value = kwargs["Title"][method]
            if value is not None:
                getattr(pstyle, method)(value)
    sheet.add_paragraph_style("Title", pstyle)

    pstyle = ParagraphStyle()
    fstyle = pstyle.get_font()
    fstyle.set_type_face(FONT_SANS_SERIF)
    fstyle.set_size(12)
    fstyle.set_bold(True)
    pstyle.set_font(fstyle)
    pstyle.set_alignment(PARA_ALIGN_LEFT)
    pstyle.set_tabs([4, 8, 12, 16])
    # Handle args:
    if "Header1" in kwargs:
        for method in kwargs["Header1"]:
            value = kwargs["Header1"][method]
            if value is not None:
                getattr(pstyle, method)(value)
    sheet.add_paragraph_style("Header1", pstyle)

    pstyle = ParagraphStyle()
    fstyle = pstyle.get_font()
    fstyle.set_type_face(FONT_SANS_SERIF)
    fstyle.set_size(10)
    fstyle.set_bold(True)
    pstyle.set_font(fstyle)
    pstyle.set_alignment(PARA_ALIGN_LEFT)
    pstyle.set_tabs([4, 8, 12, 16])
    # Handle args:
    if "Header2" in kwargs:
        for method in kwargs["Header2"]:
            value = kwargs["Header2"][method]
            if value is not None:
                getattr(pstyle, method)(value)
    sheet.add_paragraph_style("Header2", pstyle)

    pstyle = ParagraphStyle()
    fstyle = pstyle.get_font()
    fstyle.set_type_face(FONT_SANS_SERIF)
    fstyle.set_size(10)
    fstyle.set_bold(True)
    fstyle.set_italic(True)
    pstyle.set_font(fstyle)
    pstyle.set_alignment(PARA_ALIGN_LEFT)
    pstyle.set_tabs([4, 8, 12, 16])
    # Handle args:
    if "Header3" in kwargs:
        for method in kwargs["Header3"]:
            value = kwargs["Header3"][method]
            if value is not None:
                getattr(pstyle, method)(value)
    sheet.add_paragraph_style("Header3", pstyle)

    pstyle = ParagraphStyle()
    pstyle.set_tabs([4, 8, 12, 16])
    # Handle args:
    if "Normal" in kwargs:
        for method in kwargs["Normal"]:
            value = kwargs["Normal"][method]
            if value is not None:
                getattr(pstyle, method)(value)
    sheet.add_paragraph_style("Normal", pstyle)

    # Styles for tables:
    tbl = TableStyle()
    tbl.set_width(100)
    tbl.set_columns(2)
    tbl.set_column_width(0, 20)
    tbl.set_column_width(1, 80)
    # Handle args:
    if "Table" in kwargs:
        for method in kwargs["Table"]:
            value = kwargs["Table"][method]
            if value is not None:
                getattr(tbl, method)(value)
    sheet.add_table_style("Table", tbl)

    cell = TableCellStyle()
    cell.set_top_border(1)
    cell.set_bottom_border(1)
    # Handle args:
    if "TableHead" in kwargs:
        for method in kwargs["TableHead"]:
            value = kwargs["TableHead"][method]
            if value is not None:
                getattr(cell, method)(value)
    sheet.add_cell_style("TableHead", cell)

    cell = TableCellStyle()
    # Handle args:
    if "TableHeaderCell" in kwargs:
        for method in kwargs["TableHeaderCell"]:
            value = kwargs["TableHeaderCell"][method]
            if value is not None:
                getattr(cell, method)(value)
    sheet.add_cell_style("TableHeaderCell", cell)

    cell = TableCellStyle()
    cell.set_longlist(1)
    # Handle args:
    if "TableDataCell" in kwargs:
        for method in kwargs["TableDataCell"]:
            value = kwargs["TableDataCell"][method]
            if value is not None:
                getattr(cell, method)(value)
    sheet.add_cell_style("TableDataCell", cell)

    return sheet



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011  Benny Malengier
# Copyright (C) 2011  Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import MatchesFilterBase


# -------------------------------------------------------------------------
# "Sources which reference a repository by selection"
# -------------------------------------------------------------------------
class MatchesRepositoryFilter(MatchesFilterBase):
    """Citations which have a source which references the selected repository"""

    labels = [_("Repository filter name:")]
    name = _(
        "Citations with a source with a repository reference "
        "matching the <repository filter>"
    )
    description = _(
        "Matches citations with sources with a repository "
        "reference that match a certain repository filter"
    )
    category = _("General filters")

    # we want to have this filter show repository filters
    namespace = "Repository"

    def prepare(self, db, user):
        MatchesFilterBase.prepare(self, db, user)
        self.MRF_filt = self.find_filter()

    def apply(self, db, object):
        if self.MRF_filt is None:
            return False

        source_handle = object.source_handle
        source = db.get_source_from_handle(source_handle)
        repolist = [x.ref for x in source.get_reporef_list()]
        for repohandle in repolist:
            # check if repo in repository filter
            if self.MRF_filt.check(db, repohandle):
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Zsolt Foldvari
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"ToolComboEntry class."

__all__ = ["MenuItemWithData", "add_menuitem"]

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.menuitem")

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# MenuItemWithData class
#
# -------------------------------------------------------------------------


class MenuItemWithData(Gtk.MenuItem):
    """A MenuItem that stores a data property. As set_data in GTK3 is not
    working, this is a workaround to have set_data"""

    data = GObject.Property(type=object)

    def __init__(self, label=""):
        Gtk.MenuItem.__init__(self, label=label)

    def set_data(self, data):
        self.data = data

    def get_data(self, _=None):
        """obtain the data, for backward compat, we allow a dummy argument"""
        return self.data


def add_menuitem(menu, msg, obj, func):
    """
    add a menuitem to menu with label msg, which activates func, and has data
    obj
    """
    item = MenuItemWithData(label=msg)
    item.set_data(obj)
    item.connect("activate", func)
    item.show()
    menu.append(item)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
AddressBase class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .address import Address
from .const import IDENTICAL, EQUAL


# -------------------------------------------------------------------------
#
# AddressBase classes
#
# -------------------------------------------------------------------------
class AddressBase:
    """
    Base class for address-aware objects.
    """

    def __init__(self, source=None):
        """
        Initialize a AddressBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: AddressBase
        """
        self.address_list = list(map(Address, source.address_list)) if source else []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return [addr.serialize() for addr in self.address_list]

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.address_list = [Address().unserialize(item) for item in data]

    def add_address(self, address):
        """
        Add the :class:`~.address.Address` instance to the object's list of
        addresses.

        :param address: :class:`~.address.Address` instance to add to the
                        object's address list
        :type address: list
        """
        self.address_list.append(address)

    def remove_address(self, address):
        """
        Remove the specified :class:`~.address.Address` instance from the address list.

        If the instance does not exist in the list, the operation has
        no effect.

        :param address: :class:`~.address.Address` instance to remove from the
                        list
        :type address: :class:`~.address.Address`

        :returns: True if the address was removed, False if it was not in the
                  list.
        :rtype: bool
        """
        if address in self.address_list:
            self.address_list.remove(address)
            return True
        else:
            return False

    def get_address_list(self):
        """
        Return the list of :class:`~.address.Address` instances associated with
        the object.

        :returns: Returns the list of :class:`~.address.Address` instances
        :rtype: list
        """
        return self.address_list

    def set_address_list(self, address_list):
        """
        Assign the passed list to the object's list of
        :class:`~.address.Address` instances.

        :param address_list: List of :class:`~.address.Address` instances to be
                             associated with the object
        :type address_list: list
        """
        self.address_list = address_list

    def _merge_address_list(self, acquisition):
        """
        Merge the list of addresses from acquisition with our own.

        :param acquisition: the address list of this object will be merged with
                            the current address list.
        :type acquisition: AddressBase
        """
        address_list = self.address_list[:]
        for addendum in acquisition.get_address_list():
            for address in address_list:
                equi = address.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    address.merge(addendum)
                    break
            else:
                self.address_list.append(addendum)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2019-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    UpdatesPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging
from time import strftime, time, localtime

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.gen.display.place import displayer as _pd
from gramps.plugins.webreport.common import FULLCLEAR, _EVENTMAP, html_escape
from gramps.gen.lib import (
    Person,
    Family,
    Event,
    Place,
    Source,
    Repository,
    Media,
    Citation,
)
from gramps.gen.lib.date import Date

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class UpdatesPage(BasePage):
    """
    This class is responsible for displaying information about the Home page.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        ldatec = 0
        self.inc_repository = self.report.options["inc_repository"]
        self.inc_families = self.report.options["inc_families"]
        self.inc_events = self.report.options["inc_events"]
        self.inc_places = self.report.options["inc_places"]
        self.inc_sources = self.report.options["inc_sources"]
        self.inc_gallery = False

        output_file, sio = self.report.create_file("updates")
        result = self.write_header(self._("New and updated objects"))
        homepage, dummy_head, dummy_body, outerwrapper = result
        self.days = self.report.options["maxdays"]
        self.nbr = self.report.options["maxupdates"]
        cur_time = int(time())
        self.maxdays = cur_time - (86400 * self.days)

        # begin updates division
        with Html("div", class_="content", id="Updates") as section:
            outerwrapper += section
            description = self._(
                "This page contains the last updated objects "
                "in the database in the last %(days)d days and"
                " for a maximum of %(nb)d objects per object "
                "type." % {"days": self.days, "nb": self.nbr}
            )
            section += Html("p", description)

            header = self._("People")
            section += Html("h4", header)
            people = self.list_people_changed(Person)
            if people is not None:
                section += people

            if self.inc_families:
                header = self._("Families")
                section += Html("h4", header)
                families = self.list_people_changed(Family)
                if families is not None:
                    section += families

            if self.inc_events:
                header = self._("Events")
                section += Html("h4", header)
                events = self.list_people_changed(Event)
                if events is not None:
                    section += events

            if self.inc_places:
                header = self._("Places")
                section += Html("h4", header)
                places = self.list_people_changed(Place)
                if places is not None:
                    section += places

            if self.inc_sources:
                header = self._("Sources")
                section += Html("h4", header)
                sources = self.list_people_changed(Source)
                if sources is not None:
                    section += sources

            if self.inc_repository:
                header = self._("Repositories")
                section += Html("h4", header)
                repos = self.list_people_changed(Repository)
                if repos is not None:
                    section += repos

            if (
                self.report.options["gallery"]
                and not self.report.options["create_thumbs_only"]
            ):
                self.inc_gallery = True
                header = self._("Media")
                section += Html("h4", header)
                media = self.list_people_changed(Media)
                if media is not None:
                    section += media

            header = self._("Notes")
            section += Html("h4", header)
            events = self.list_notes()
            if events is not None:
                section += events

        # create clear line for proper styling
        # create footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(homepage, output_file, sio, ldatec)

    def list_notes(self):
        """
        List all notes with last change date
        """
        nb_items = 0
        section = ""

        def sort_on_change(handle):
            """sort records based on the last change time"""
            fct = self.report.database.get_note_from_handle
            obj = fct(handle)
            timestamp = obj.get_change_time()
            return timestamp

        note_list = self.report.database.get_note_handles()
        obj_list = sorted(note_list, key=sort_on_change, reverse=True)
        with Html("table", class_="list", id="list") as section:
            for handle in obj_list:
                show = False
                date = obj = None
                obj = self.report.database.get_note_from_handle(handle)
                if obj:
                    text = html_escape(obj.get()[:50])
                    timestamp = obj.get_change_time()
                    if timestamp - self.maxdays > 0:
                        handle_list = set(
                            self.report.database.find_backlink_handles(
                                handle,
                                include_classes=[
                                    "Person",
                                    "Family",
                                    "Event",
                                    "Place",
                                    "Media",
                                    "Source",
                                    "Citation",
                                    "Repository",
                                ],
                            )
                        )
                        tims = localtime(timestamp)
                        odat = Date(tims.tm_year, tims.tm_mon, tims.tm_mday)
                        date = self.rlocale.date_displayer.display(odat)
                        date += strftime(" %X", tims)
                        if handle_list:
                            srbd = self.report.database
                            srbkref = self.report.bkref_dict
                            for obj_t, r_handle in handle_list:
                                if obj_t == "Person":
                                    if r_handle in srbkref[Person]:
                                        name = self.new_person_link(r_handle)
                                        show = True
                                elif obj_t == "Family":
                                    if r_handle in srbkref[Family]:
                                        fam = srbd.get_family_from_handle(r_handle)
                                        fam = self._("Family")
                                        name = self.family_link(r_handle, fam)
                                        if self.inc_families:
                                            show = True
                                elif obj_t == "Place":
                                    if r_handle in srbkref[Place]:
                                        plc = srbd.get_place_from_handle(r_handle)
                                        plcn = _pd.display(self.report.database, plc)
                                        name = self.place_link(r_handle, plcn)
                                        if self.inc_places:
                                            show = True
                                elif obj_t == "Event":
                                    if r_handle in srbkref[Event]:
                                        evt = srbd.get_event_from_handle(r_handle)
                                        evtn = self._(evt.get_type().xml_str())
                                        name = self.event_link(r_handle, evtn)
                                        if self.inc_events:
                                            show = True
                                elif obj_t == "Media":
                                    if r_handle in srbkref[Media]:
                                        media = srbd.get_media_from_handle(r_handle)
                                        evtn = media.get_description()
                                        name = self.media_link(
                                            r_handle, evtn, evtn, usedescr=False
                                        )
                                        if self.inc_gallery:
                                            show = True
                                elif obj_t == "Citation":
                                    if r_handle in srbkref[Citation]:
                                        cit = srbd.get_event_from_handle(r_handle)
                                        citsrc = cit.source_handle
                                        evtn = self._("Citation")
                                        name = self.source_link(citsrc, evtn)
                                        if self.inc_sources:
                                            show = True
                                elif obj_t == "Source":
                                    if r_handle in srbkref[Source]:
                                        src = srbd.get_source_from_handle(r_handle)
                                        evtn = src.get_title()
                                        name = self.source_link(r_handle, evtn)
                                        if self.inc_sources:
                                            show = True
                                elif obj_t == "Repository":
                                    if r_handle in srbkref[Repository]:
                                        rep = srbd.get_repository_from_handle(r_handle)
                                        evtn = rep.get_name()
                                        name = self.repository_link(r_handle, evtn)
                                        if self.inc_repository:
                                            show = True
                        if show:
                            row = Html("tr")
                            section += row
                            row += Html("td", date, class_="date")
                            row += Html("td", text)
                            row += Html("td", name)
                            nb_items += 1
                            if nb_items > self.nbr:
                                break
        return section

    def list_people_changed(self, object_type):
        """
        List all records with last change date
        """
        nb_items = 0

        def sort_on_change(handle):
            """sort records based on the last change time"""
            try:
                obj = fct(handle)
                timestamp = obj.get_change_time()
            except exception:
                print("In sort :", handle, " not found for", object_type)
                timestamp = 0
            return timestamp

        if object_type == Person:
            fct = self.report.database.get_person_from_handle
            fct_link = self.new_person_link
        elif object_type == Family:
            fct = self.report.database.get_family_from_handle
            fct_link = self.family_link
        elif object_type == Event:
            fct = self.report.database.get_event_from_handle
        elif object_type == Place:
            fct = self.report.database.get_place_from_handle
            fct_link = self.place_link
        elif object_type == Source:
            fct = self.report.database.get_source_from_handle
            fct_link = self.source_link
        elif object_type == Repository:
            fct = self.report.database.get_repository_from_handle
            fct_link = self.repository_link
        elif object_type == Media:
            fct = self.report.database.get_media_from_handle
        obj_list = sorted(
            self.report.obj_dict[object_type], key=sort_on_change, reverse=True
        )
        with Html("table", class_="list", id="list") as section:
            for handle in obj_list:
                date = obj = None
                name = ""
                try:
                    obj = fct(handle)
                except exception:
                    print("Handle not found:", handle, " not found for", object_type)
                    continue
                if object_type == Person:
                    name = fct_link(handle)
                elif object_type == Family:
                    name = self.report.get_family_name(obj)
                    name = fct_link(handle, name)
                elif object_type == Event:
                    otype = obj.get_type()
                    date = obj.get_date_object()
                    if int(otype) in _EVENTMAP:
                        handle_list = set(
                            self.report.database.find_backlink_handles(
                                handle, include_classes=["Family", "Person"]
                            )
                        )
                    else:
                        handle_list = set(
                            self.report.database.find_backlink_handles(
                                handle, include_classes=["Person"]
                            )
                        )
                    if handle_list:
                        name = Html("span", self._(otype.xml_str()) + " ")
                        for obj_t, r_handle in handle_list:
                            if obj_t == "Person":
                                name += self.new_person_link(r_handle)
                            else:
                                srbd = self.report.database
                                fam = srbd.get_family_from_handle(r_handle)
                                srgfn = self.report.get_family_name
                                name += self.family_link(r_handle, srgfn(fam))
                elif object_type == Place:
                    name = _pd.display(self.report.database, obj)
                    name = fct_link(handle, name)
                elif object_type == Source:
                    name = obj.get_title()
                    name = fct_link(handle, name)
                elif object_type == Repository:
                    name = obj.get_name()
                    name = fct_link(handle, name)
                elif object_type == Media:
                    name = obj.get_description()
                    url = self.report.build_url_fname_html(handle, "img")
                    name = Html("a", name, href=url, title=name)
                if obj:
                    timestamp = obj.get_change_time()
                    if timestamp - self.maxdays > 0:
                        nb_items += 1
                        if nb_items > self.nbr:
                            break
                        tims = localtime(timestamp)
                        odat = Date(tims.tm_year, tims.tm_mon, tims.tm_mday)
                        date = self.rlocale.date_displayer.display(odat)
                        date += strftime(" %X", tims)
                        row = Html("tr")
                        section += row
                        row += Html("td", date, class_="date")
                        row += Html("td", name)
        return section



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# gen.filters.rules/Repository/_ChangedSince.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._changedsincebase import ChangedSinceBase


# -------------------------------------------------------------------------
#
# ChangedSince
#
# -------------------------------------------------------------------------
class ChangedSince(ChangedSinceBase):
    """Rule that checks for repositories changed since a specific time."""

    labels = [_("Changed after:"), _("but before:")]
    name = _("Repositories changed after <date time>")
    description = _(
        "Matches repository records changed after a specified "
        "date/time (yyyy-mm-dd hh:mm:ss) or in the range, if a second "
        "date/time is given."
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the base class for GRAMPS' DataView classes
"""

# ----------------------------------------------------------------
#
# python modules
#
# ----------------------------------------------------------------
from abc import ABCMeta, abstractmethod
import logging

_LOG = logging.getLogger(".pageview")

# ----------------------------------------------------------------
#
# gtk
#
# ----------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ----------------------------------------------------------------
#
# Gramps
#
# ----------------------------------------------------------------
from gramps.gen.errors import WindowActiveError
from ..dbguielement import DbGUIElement
from ..widgets.grampletbar import GrampletBar
from ..configure import ConfigureDialog
from gramps.gen.config import config
from ..uimanager import ActionGroup


# ------------------------------------------------------------------------------
#
# PageView
#
# ------------------------------------------------------------------------------
class PageView(DbGUIElement, metaclass=ABCMeta):
    """
    The PageView class is the base class for all Data Views in Gramps.  All
    Views should derive from this class. The ViewManager understands the public
    interface of this class

    The following attributes are available
    ..attribute:: active
      is the view active at the moment (visible in Gramps as the main view)
    ..attribute:: dirty
      bool, is the view current with the database or not. A dirty view triggers
      a rebuild when it becomes active
    ..attribute:: _dirty_on_change_inactive
      read/write bool by inheriting classes.
      Views can behave two ways to signals:
      1. if the view is inactive, set it dirty, and rebuild the view when
          it becomes active. In this case, this method returns True
      2. if the view is inactive, try to stay in sync with database. Only
         rebuild or other large changes make view dirty
    ..attribute:: title
      title of the view
    ..attribute:: dbstate
      the dbstate
    ..attribute:: uistate
      the displaystate
    ..attribute:: category
      the category to which the view belongs. Views in the same category are
      placed behind the same button in the sidebar
    """

    CONFIGSETTINGS = []

    def __init__(self, title, pdata, dbstate, uistate):
        self.title = title
        self.pdata = pdata
        self.dbstate = dbstate
        self.uistate = uistate
        self.action_list = []
        self.action_toggle_list = []
        self.action_group = None
        self.additional_action_groups = []
        self.additional_uis = []
        self.ui_def = [
            """
          <placeholder id="Bars">
            <item>
              <attribute name="action">win.Sidebar</attribute>
              <attribute name="label" translatable="yes">_Sidebar</attribute>
            </item>
             <item>
              <attribute name="action">win.Bottombar</attribute>
              <attribute name="label" translatable="yes">_Bottombar</attribute>
            </item>
          </placeholder>
            """
        ]
        self.dirty = True
        self.active = False
        self._dirty_on_change_inactive = True

        if isinstance(self.pdata.category, tuple):
            self.category, self.translated_category = self.pdata.category
        else:
            raise AttributeError("View category must be (name, translated-name)")
        self.ident = self.category + "_" + self.pdata.id

        self.dbstate.connect("no-database", self.disable_action_group)
        self.dbstate.connect("database-changed", self.enable_action_group)
        self.uistate.window.connect("key-press-event", self.key_press_handler)

        self.model = None
        self.selection = None
        self.handle_col = 0

        self._config = None
        self.init_config()

        self.filter_class = None
        self.top = None
        self.sidebar = None
        self.bottombar = None
        self.widget = None
        self.vpane = None

        DbGUIElement.__init__(self, dbstate.db)

    def build_interface(self):
        """
        Builds the container widget for the interface.
        Returns a gtk container widget.
        """
        defaults = self.get_default_gramplets()
        self.sidebar = GrampletBar(
            self.dbstate,
            self.uistate,
            self,
            self.ident + "_sidebar",
            defaults[0],
            Gtk.Orientation.VERTICAL,
        )
        self.bottombar = GrampletBar(
            self.dbstate,
            self.uistate,
            self,
            self.ident + "_bottombar",
            defaults[1],
            Gtk.Orientation.HORIZONTAL,
        )
        hpane = Gtk.Paned()
        self.vpane = Gtk.Paned(orientation=Gtk.Orientation.VERTICAL)
        hpane.pack1(self.vpane, resize=True, shrink=False)
        hpane.pack2(self.sidebar, resize=False, shrink=False)
        hpane.show()
        self.vpane.show()

        self.widget = self.build_widget()
        self.widget.show_all()
        self.widget.set_name("view")
        self.vpane.pack1(self.widget, resize=True, shrink=False)
        self.vpane.pack2(self.bottombar, resize=False, shrink=False)
        self.vpane.show()
        self._setup_slider_config(self.vpane, "vpane.slider-position", position=-1)

        self.sidebar_toggled(self.sidebar.get_property("visible"))
        self.hpane_sig = hpane.connect("draw", self.set_page_slider)

        return hpane

    def set_page_slider(self, widget, dummy):
        """Setup slider.  We have the page realized at this point."""
        widget.disconnect(self.hpane_sig)
        # get current width of pane
        width = widget.get_allocated_width()
        # default will use natural size for sidebar until it gets to 400 pix
        side_ch = self.sidebar.get_children()  # Gtk Notebook
        try:
            vp_ch = side_ch[0].get_children()  # Gtk Viewport child
            ch_width = vp_ch[0].get_preferred_width()[0] + 3
        except AttributeError:
            ch_width = 300  # needed if no Gramplet installed
        pos = width - min(ch_width, 400)
        self._setup_slider_config(widget, "hpane.slider-position", position=pos)

    def _setup_slider_config(self, widget, setting, position=-1):
        """
        Setup the slider configuration setting.
        """
        self._config.register(setting, position)
        widget.set_position(self._config.get(setting))
        widget.connect("notify::position", self._position_changed, setting)

    def _position_changed(self, widget, position, setting):
        """
        Called when the position property of the pane is changed.
        """
        self._config.set(setting, widget.get_position())

    def __sidebar_toggled(self, action, value):
        """
        Called when the sidebar is toggled.
        """
        action.set_state(value)  # change GUI
        if value.get_boolean():
            self.sidebar.show()
            self.sidebar_toggled(True)
        else:
            self.sidebar.hide()
            self.sidebar_toggled(False)

    def __bottombar_toggled(self, action, value):
        """
        Called when the bottombar is toggled.
        """
        action.set_state(value)  # change GUI
        if value.get_boolean():
            self.bottombar.show()
        else:
            self.bottombar.hide()

    def sidebar_toggled(self, active):
        """
        Called when the sidebar is toggled.
        """
        pass

    def get_default_gramplets(self):
        """
        Get the default gramplets for the Gramps sidebar and bottombar.
        Returns a 2-tuple.  The first element is a tuple of sidebar gramplets
        and the second element is a tuple of bottombar gramplets.

        Views should override this method to define default gramplets.
        """
        return ((), ())

    def key_press_handler(self, widget, event):
        """
        A general keypress handler. Override if you want to handle
        special control characters, like control+c (copy) or control+v
        (paste).
        """
        return False

    def copy_to_clipboard(self, objclass, handles):
        """
        This code is called on Control+C in a navigation view. If the
        copy can be handled, it returns true, otherwise false.

        The code brings up the Clipboard (if already exists) or
        creates it. The copy is handled through the drag and drop
        system.
        """
        import pickle
        from ..clipboard import ClipboardWindow, obj2target

        handled = False
        for handle in handles:
            if handle is None:
                continue
            clipboard = None
            for widget in self.uistate.gwm.window_tree:
                if isinstance(widget, ClipboardWindow):
                    clipboard = widget
            if clipboard is None:
                clipboard = ClipboardWindow(self.dbstate, self.uistate)
            # Construct a drop:
            drag_type = obj2target(objclass)
            if drag_type:

                class Selection:
                    def __init__(self, data):
                        self.data = data

                    def get_data(self):
                        return self.data

                class Context:
                    targets = [Gdk.atom_intern(drag_type.name(), False)]
                    action = 1

                    def list_targets(self):
                        return Context.targets

                    def get_actions(self):
                        return Context.action

                # eg: ('person-link', 23767, '27365123671', 0)
                data = (drag_type.name(), id(self), handle, 0)
                clipboard.object_list.object_drag_data_received(
                    clipboard.object_list._widget,  # widget
                    Context(),  # drag type and action
                    0,
                    0,  # x, y
                    Selection(pickle.dumps(data)),  # pickled data
                    None,  # info (not used)
                    -1,
                )  # time
                handled = True
        return handled

    def call_paste(self):
        """
        This code is called on Control+V in a navigation view. If the
        copy can be handled, it returns true, otherwise false.

        The code creates the Clipboard if it does not already exist.
        """
        from ..clipboard import ClipboardWindow

        clipboard = None
        for widget in self.uistate.gwm.window_tree:
            if isinstance(widget, ClipboardWindow):
                clipboard = widget
        if clipboard is None:
            clipboard = ClipboardWindow(self.dbstate, self.uistate)
            return True
        return False

    def post(self):
        """
        Called after a page is created.
        """
        pass

    def set_active(self):
        """
        Called with the PageView is set as active. If the page is "dirty",
        then we rebuild the data.
        """
        self.sidebar.set_active()
        self.bottombar.set_active()
        self.active = True
        new_title = "%s - %s - Gramps" % (
            self.dbstate.db.get_dbname(),
            self.get_title(),
        )
        self.uistate.window.set_title(new_title)
        if self.dirty:
            self.uistate.set_busy_cursor(True)
            self.build_tree()
            self.uistate.set_busy_cursor(False)

    def set_inactive(self):
        """
        Marks page as being inactive (not currently displayed)
        """
        self.sidebar.set_inactive()
        self.bottombar.set_inactive()
        self.active = False

    def post_create(self):
        if self.vpane:
            self._setup_slider_config(self.vpane, "vpane.slider-position")
            self.vpane = None

    @abstractmethod
    def build_tree(self):
        """
        Rebuilds the current display. This must be overridden by the derived
        class.
        """

    def ui_definition(self):
        """
        returns the XML UI definition for the UIManager
        """
        return self.ui_def

    def additional_ui_definitions(self):
        """
        Return any additional interfaces for the UIManager that the view
        needs to define.
        """
        return self.additional_uis

    def disable_action_group(self):
        """
        Turns off the visibility of the View's action group, if defined
        """
        if self.action_group:
            self.uistate.uimanager.set_actions_visible(self.action_group, False)

    def enable_action_group(self, obj):
        """
        Turns on the visibility of the View's action group, if defined
        """
        if self.action_group:
            self.uistate.uimanager.set_actions_visible(self.action_group, True)

    def get_stock(self):
        """
        Return image associated with the view category, which is used for the
        icon for the button.
        """
        return "image-missing"

    def get_viewtype_stock(self):
        """
        Return immage associated with the viewtype inside a view category, it
        will be used for the icon on the button to select view in the category
        """
        return "image-missing"

    def get_title(self):
        """
        Return the title of the view. This is used to define the text for the
        button, and for the tab label.
        """
        return self.title

    def get_category(self):
        """
        Return the category name of the view. This is used to define
        the text for the button, and for the tab label.
        """
        return self.category

    def get_translated_category(self):
        """
        Return the translated category name of the view. This is used
        to define the text for the button, and for the tab label.
        """
        return self.translated_category

    def get_display(self):
        """
        Builds the graphical display, returning the top level widget.
        """
        if not self.top:
            self.top = self.build_interface()
        return self.top

    @abstractmethod
    def build_widget(self):
        """
        Builds the container widget for the main view pane. Must be overridden
        by the base class. Returns a gtk container widget.
        """

    def define_actions(self):
        """
        Defines the UIManager actions. Called by the ViewManager to set up the
        View. The user typically defines self.action_list and
        self.action_toggle_list in this function.
        """
        self._add_toggle_action(
            "Sidebar",
            self.__sidebar_toggled,
            "<shift><PRIMARY>R",
            self.sidebar.get_property("visible"),
        )
        self._add_toggle_action(
            "Bottombar",
            self.__bottombar_toggled,
            "<shift><PRIMARY>B",
            self.bottombar.get_property("visible"),
        )

    def __build_action_group(self):
        """
        Create an UIManager ActionGroup from the values in self.action_list
        and self.action_toggle_list. The user should define these in
        self.define_actions
        """
        self.action_group = ActionGroup(name=self.title)
        if len(self.action_list) > 0:
            self.action_group.add_actions(self.action_list)
        if len(self.action_toggle_list) > 0:
            self.action_group.add_actions(self.action_toggle_list)

    def _add_action(self, name, callback=None, accel=None):
        """
        Add an action to the action list for the current view.
        """
        self.action_list.append((name, callback, accel))

    def _add_toggle_action(self, name, callback=None, accel=None, value=False):
        """
        Add a toggle action to the action list for the current view.
        """
        self.action_toggle_list.append((name, callback, accel, value))

    def get_actions(self):
        """
        Return the actions that should be used for the view. This includes the
        standard action group (which handles the main toolbar), along with
        additional action groups.

        If the action group is not defined, we build it the first time. This
        allows us to delay the intialization until it is really needed.

        The ViewManager uses this function to extract the actions to install
        into the UIManager.
        """
        if not self.action_group:
            self.__build_action_group()
        return [self.action_group] + self.additional_action_groups

    def _add_action_group(self, group):
        """
        Allows additional action groups to be added to the view.
        """
        self.additional_action_groups.append(group)

    def change_page(self):
        """
        Called when the page changes.
        """
        self.uistate.clear_filter_results()

    def on_delete(self):
        """
        Method called on shutdown. Data views should put code here
        that should be called when quiting the main application.
        """
        self.sidebar.on_delete()
        self.bottombar.on_delete()
        self._config.save()

    def init_config(self):
        """
        If you need a view with a config, then call this method in the
        build_widget or __init__ method. It will set up a config file for the
        view, and use CONFIGSETTINGS to set the config defaults.
        The config is later accessbile via self._config
        So you can do
        self._config.get("section.variable1")
        self._config.set("section.variable1", value)
        self._config.save()

        CONFIGSETTINGS should be a list with tuples like
        ("section.variable1", value)
        """
        if self._config:
            return
        self._config = config.register_manager(self.ident, use_config_path=True)
        for section, value in self.CONFIGSETTINGS:
            self._config.register(section, value)
        self._config.init()
        self.config_connect()

    def config_connect(self):
        """
        Overwrite this method to set connects to the config file to monitor
        changes. This method will be called after the ini file is initialized
        Eg:
            self.config.connect("section.option", self.callback)
        """
        pass

    def config_callback(self, callback):
        """
        Convenience wrappen to create a callback for a config setting
        :param callback: a callback function to call.
        """
        return lambda arg1, arg2, arg3, arg4: callback()

    def can_configure(self):
        """
        Inheriting classes should set if the view has a configure window or not
        :return: bool
        """
        return False

    def _get_configure_page_funcs(self):
        """
        Return a list of functions that create gtk elements to use in the
        notebook pages of the Configure view

        :return: list of functions
        """
        return []

    def configure(self):
        """
        Open the configure dialog for the view.
        """
        title = _("Configure %(cat)s - %(view)s") % {
            "cat": self.get_translated_category(),
            "view": self.get_title(),
        }

        if self.can_configure():
            config_funcs = self._get_configure_page_funcs()
        else:
            config_funcs = []
        if self.sidebar:
            config_funcs += self.sidebar.get_config_funcs()
        if self.bottombar:
            config_funcs += self.bottombar.get_config_funcs()

        try:
            ViewConfigureDialog(
                self.uistate,
                self.dbstate,
                config_funcs,
                self,
                self._config,
                dialogtitle=title,
                ident=_("%(cat)s - %(view)s")
                % {"cat": self.get_translated_category(), "view": self.get_title()},
            )
        except WindowActiveError:
            return


class ViewConfigureDialog(ConfigureDialog):
    """
    All views can have their own configuration dialog
    """

    def __init__(
        self,
        uistate,
        dbstate,
        configure_page_funcs,
        configobj,
        configmanager,
        dialogtitle=_("Preferences"),
        on_close=None,
        ident="",
    ):
        self.ident = ident
        ConfigureDialog.__init__(
            self,
            uistate,
            dbstate,
            configure_page_funcs,
            configobj,
            configmanager,
            dialogtitle=dialogtitle,
            on_close=on_close,
        )
        self.setup_configs("interface.viewconfiguredialog", 420, 500)

    def build_menu_names(self, obj):
        return (_("Configure %s View") % self.ident, None)


class DummyPage(PageView):
    """
    A Dummy page for testing or errors
    """

    def __init__(self, title, pdata, dbstate, uistate, msg1="", msg2=""):
        self.msg = msg1
        self.msg2 = msg2
        PageView.__init__(self, title, pdata, dbstate, uistate)

    def build_widget(self):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=1)
        # top widget at the top
        box.pack_start(
            Gtk.Label(
                label=_("View %(name)s: %(msg)s")
                % {"name": self.title, "msg": self.msg}
            ),
            False,
            False,
            0,
        )
        tv = Gtk.TextView()
        tb = tv.get_buffer()
        tb.insert_at_cursor(self.msg2)
        box.pack_start(tv, False, False, 0)
        return box

    def build_tree(self):
        pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2006  Donald N. Allingham
# Copyright (C) 2014       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide an interface to the gspell interface. If the gspell package is not
present, we default to no spell checking.

"""

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Set up logging
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".Spell")

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

HAVE_GSPELL = False

try:
    import gi

    gi.require_version("Gspell", "1")
    from gi.repository import Gspell

    langs = Gspell.language_get_available()
    for lang in langs:
        LOG.debug("%s (%s) dict available", lang.get_name(), lang.get_code())
    if langs:
        HAVE_GSPELL = True
    else:
        LOG.warning(_("You have no installed dictionaries."))
except (ImportError, ValueError):
    pass

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.config import config

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------


class Spell:
    """Attach a Gspell instance to the passed TextView instance."""

    _spellcheck_options = {"off": _("Off")}

    if HAVE_GSPELL:
        _spellcheck_options["on"] = _("On")

    def __init__(self, textview):
        self._active_spellcheck = "off"

        if not HAVE_GSPELL:
            return

        locale_code = glocale.locale_code()
        gspell_language = None
        if locale_code is not None:
            gspell_language = Gspell.language_lookup(locale_code[:5])
            if gspell_language is None:
                gspell_language = Gspell.language_lookup(locale_code[:2])
        checker = Gspell.Checker.new(gspell_language)
        buffer = Gspell.TextBuffer.get_from_gtk_text_buffer(textview.get_buffer())
        buffer.set_spell_checker(checker)
        self.gspell_view = Gspell.TextView.get_from_gtk_text_view(textview)

        if config.get("behavior.spellcheck"):
            self.spellcheck = "on"
        else:
            self.spellcheck = "off"

        self.__real_set_active_spellcheck(self.spellcheck)

    # Private

    def __real_set_active_spellcheck(self, spellcheck_code):
        """Set active spellcheck by its code."""
        if self._active_spellcheck == spellcheck_code:
            return

        self.gspell_view.set_inline_spell_checking(spellcheck_code == "on")
        self.gspell_view.set_enable_language_menu(spellcheck_code == "on")
        self._active_spellcheck = spellcheck_code

    # Public API

    def get_all_spellchecks(self):
        """Get the list of installed spellcheck names."""
        return list(self._spellcheck_options.values())

    def set_active_spellcheck(self, spellcheck):
        """Set active spellcheck by it's name."""
        for code, name in list(self._spellcheck_options.items()):
            if name == spellcheck:
                self.__real_set_active_spellcheck(code)
                return

    def get_active_spellcheck(self):
        """Get the name of the active spellcheck."""
        return self._spellcheck_options[self._active_spellcheck]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Gerald Britton <gerald.britton@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Glade
#
# ------------------------------------------------------------------------

"""
Glade file operations

This module exports the Glade class.

"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import sys
import os
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GLADE_DIR, GRAMPS_LOCALE as glocale
from gramps.gen.constfunc import is_quartz

# ------------------------------------------------------------------------
#
# Glade class. Derived from Gtk.Builder
#
# ------------------------------------------------------------------------


class Glade(Gtk.Builder):
    """
    Glade class: Manage glade files as Gtk.Builder objects
    """

    __slots__ = ["__toplevel", "__filename", "__dirname"]

    def __init__(self, filename=None, dirname=None, toplevel=None, also_load=[]):
        """
        Class Constructor: Returns a new instance of the Glade class

        :type  filename: string or None
        :param filename: The name of the glade file to be used. Defaults to None
        :type  dirname: string or None
        :param dirname: The directory to search for the glade file. Defaults to
                    None which will cause a search for the file in the default
                    directory followed by the directory of the calling module.
        :type  toplevel: string or None
        :param toplevel: The toplevel object to search for in the glade file.
                     Defaults to None, which will cause a search for a toplevel
                     matching the supplied name.
        :type  also_load: list of strings
        :param also_load: Additional toplevel objects to load from the glade
                     file.  These are typically liststore or other objects
                     needed to operate the toplevel object.
                     Defaults to [] (empty list), which will not load
                     additional objects.
        :rtype:   object reference
        :returns:  reference to the newly-created Glade instance

        This operates in two modes; when no toplevel parameter is supplied,
        the entire Glade file is loaded. It is the responsibility of the user
        to make sure ALL toplevel objects are destroyed.

        When a toplevel parameter is supplied, only that object and any
        additional objects requested in the also_load parameter are loaded.
        The user only has to destroy the requested toplevel objects.
        """
        Gtk.Builder.__init__(self)
        self.set_translation_domain(glocale.get_localedomain())

        filename_given = filename is not None
        dirname_given = dirname is not None

        # if filename not given, use module name to derive it

        if not filename_given:
            filename = sys._getframe(1).f_code.co_filename
            filename = os.path.basename(filename)
            filename = filename.rpartition(".")[0] + ".glade"
            filename = filename.lstrip("_").lower()

        # if dirname not given, use current directory

        if not dirname_given:
            dirname = sys._getframe(1).f_code.co_filename
            dirname = os.path.dirname(dirname)

        # try to find the glade file

        if filename_given and dirname_given:  # both given -- use them
            path = os.path.join(dirname, filename)

        elif filename_given:  # try default directory first
            path = os.path.join(GLADE_DIR, filename)
            if not os.path.exists(path):  # then module directory
                path = os.path.join(dirname, filename)

        elif dirname_given:  # dirname given -- use it
            path = os.path.join(dirname, filename)

        # neither filename nor dirname given.  Try:
        # 1. derived filename in default directory
        # 2. derived filename in module directory

        else:
            path = os.path.join(GLADE_DIR, filename)
            if not os.path.exists(path):
                path = os.path.join(dirname, filename)

        # try to build Gtk objects from glade file.  Let exceptions happen

        self.__dirname, self.__filename = os.path.split(path)

        # try to find the toplevel widget

        # toplevel is given
        if toplevel:
            loadlist = [toplevel] + also_load
            with open(path, "r", encoding="utf-8") as builder_file:
                data = builder_file.read()
                if is_quartz():
                    data = data.replace("GDK_CONTROL_MASK", "GDK_META_MASK")
                self.add_objects_from_string(data, loadlist)
            self.__toplevel = self.get_object(toplevel)
        # toplevel not given
        else:
            with open(path, "r", encoding="utf-8") as builder_file:
                data = builder_file.read()
                if is_quartz():
                    data = data.replace("GDK_CONTROL_MASK", "GDK_META_MASK")
                self.add_from_string(data)
            # first, use filename as possible toplevel widget name
            self.__toplevel = self.get_object(filename.rpartition(".")[0])

            # next try lowercase filename as possible widget name
            if not self.__toplevel:
                self.__toplevel = self.get_object(filename.rpartition(".")[0].lower())

                if not self.__toplevel:
                    # if no match found, search for first toplevel widget
                    for obj in self.get_objects():
                        if hasattr(obj, "get_toplevel"):
                            self.__toplevel = obj.get_toplevel()
                            break
                    else:
                        self.__toplevel = None

    def __get_filename(self):
        """
        __get_filename: return filename of glade file
        :rtype:   string
        :returns:  filename of glade file
        """
        return self.__filename

    filename = property(__get_filename)

    def __get_dirname(self):
        """
        __get_dirname: return directory where glade file found
        :rtype:   string
        :returns:  directory where glade file found
        """
        return self.__dirname

    dirname = property(__get_dirname)

    def __get_toplevel(self):
        """
        __get_toplevel: return toplevel object
        :rtype:   object
        :returns:  toplevel object
        """
        return self.__toplevel

    def __set_toplevel(self, toplevel):
        """
        __set_toplevel: set toplevel object

        :type  toplevel: string
        :param toplevel: The name of the toplevel object to use
        """
        self.__toplevel = self.get_object(toplevel)

    toplevel = property(__get_toplevel, __set_toplevel)

    def get_child_object(self, value, toplevel=None):
        """
        get_child_object: search for a child object, by name, within a given
                          toplevel.  If no toplevel argument is supplied, use
                          the toplevel attribute for this instance
        :type  value: string
        :param value: The name of the child object to find
        :type  toplevel: string
        :param toplevel: The name of the toplevel object to us
        :rtype:   object
        :returns:  child object
        """
        if not toplevel:
            toplevel = self.__toplevel
            if not toplevel:
                raise ValueError("Top level object required")

        if isinstance(toplevel, str):
            toplevel = self.get_object(toplevel)

        # Simple Breadth-First Search

        queue = [toplevel]
        while queue:
            obj = queue.pop(0)
            obj_id = Gtk.Buildable.get_name(obj)
            if obj_id == value:
                return obj
            if hasattr(obj, "get_children"):
                queue += obj.get_children()

        return None



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013  Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Date strings to translate per each language for display and parsing.

__main__
--------

Run this code with the appropriate ``LANG`` and ``LC_DATE`` set for your target
language, in order to generate the .po snippets initialized with the strings
from your locale (from the deprecated data provided in _grampslocale).

E.g., for French::

    LANG=fr_FR.utf8 LC_ALL=fr_FR.utf8 GRAMPS_RESOURCES=$PWD python -m gramps.gen.datehandler._datestrings

Then merge the output into your language's .po file, and further modify the
strings as needed. Then remove the strings from your language's
:class:`DateParserXX` and :class:`DateHandlerXX` classes.
"""

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".DateStrings")


# -------------------------------------------------------------------------
#
# DateStrings
#
# -------------------------------------------------------------------------
class DateStrings:
    """
    String tables for :class:`.DateDisplay` and :class:`.DateParser`.
    """

    # This table needs not be localized, it's only for parsing
    # Swedish calendar dates using Swedish month names.
    # Display of these months uses the regular long_months.
    # TODO should we pack these into alt_long_months instead?
    swedish_SV = (
        "",
        "Januari",
        "Februari",
        "Mars",
        "April",
        "Maj",
        "Juni",
        "Juli",
        "Augusti",
        "September",
        "Oktober",
        "November",
        "December",
    )

    def __init__(self, locale):
        _ = locale.translation.lexgettext

        self.long_months = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("January", "localized lexeme inflections"),
            _("February", "localized lexeme inflections"),
            _("March", "localized lexeme inflections"),
            _("April", "localized lexeme inflections"),
            _("May", "localized lexeme inflections"),
            _("June", "localized lexeme inflections"),
            _("July", "localized lexeme inflections"),
            _("August", "localized lexeme inflections"),
            _("September", "localized lexeme inflections"),
            _("October", "localized lexeme inflections"),
            _("November", "localized lexeme inflections"),
            _("December", "localized lexeme inflections"),
        )

        self.short_months = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("Jan", "localized lexeme inflections - short month form"),
            _("Feb", "localized lexeme inflections - short month form"),
            _("Mar", "localized lexeme inflections - short month form"),
            _("Apr", "localized lexeme inflections - short month form"),
            _("May", "localized lexeme inflections - short month form"),
            _("Jun", "localized lexeme inflections - short month form"),
            _("Jul", "localized lexeme inflections - short month form"),
            _("Aug", "localized lexeme inflections - short month form"),
            _("Sep", "localized lexeme inflections - short month form"),
            _("Oct", "localized lexeme inflections - short month form"),
            _("Nov", "localized lexeme inflections - short month form"),
            _("Dec", "localized lexeme inflections - short month form"),
        )

        _ = locale.translation.sgettext
        self.alt_long_months = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to add proper alternatives to be recognized in your localized
            # DateParser code!
            _("", "alternative month names for January"),
            _("", "alternative month names for February"),
            _("", "alternative month names for March"),
            _("", "alternative month names for April"),
            _("", "alternative month names for May"),
            _("", "alternative month names for June"),
            _("", "alternative month names for July"),
            _("", "alternative month names for August"),
            _("", "alternative month names for September"),
            _("", "alternative month names for October"),
            _("", "alternative month names for November"),
            _("", "alternative month names for December"),
        )

        self.calendar = (
            # Must appear in the order indexed by Date.CAL_... numeric constants
            _("Gregorian", "calendar"),
            _("Julian", "calendar"),
            _("Hebrew", "calendar"),
            _("French Republican", "calendar"),
            _("Persian", "calendar"),
            _("Islamic", "calendar"),
            _("Swedish", "calendar"),
        )
        _ = locale.translation.lexgettext

        self.hebrew = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("Tishri", "Hebrew month lexeme"),
            _("Heshvan", "Hebrew month lexeme"),
            _("Kislev", "Hebrew month lexeme"),
            _("Tevet", "Hebrew month lexeme"),
            _("Shevat", "Hebrew month lexeme"),
            _("AdarI", "Hebrew month lexeme"),
            _("AdarII", "Hebrew month lexeme"),
            _("Nisan", "Hebrew month lexeme"),
            _("Iyyar", "Hebrew month lexeme"),
            _("Sivan", "Hebrew month lexeme"),
            _("Tammuz", "Hebrew month lexeme"),
            _("Av", "Hebrew month lexeme"),
            _("Elul", "Hebrew month lexeme"),
        )

        self.french = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("Vendémiaire", "French month lexeme"),
            _("Brumaire", "French month lexeme"),
            _("Frimaire", "French month lexeme"),
            _("Nivôse", "French month lexeme"),
            _("Pluviôse", "French month lexeme"),
            _("Ventôse", "French month lexeme"),
            _("Germinal", "French month lexeme"),
            _("Floréal", "French month lexeme"),
            _("Prairial", "French month lexeme"),
            _("Messidor", "French month lexeme"),
            _("Thermidor", "French month lexeme"),
            _("Fructidor", "French month lexeme"),
            _("Extra", "French month lexeme"),
        )

        self.islamic = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("Muharram", "Islamic month lexeme"),
            _("Safar", "Islamic month lexeme"),
            _("Rabi`al-Awwal", "Islamic month lexeme"),
            _("Rabi`ath-Thani", "Islamic month lexeme"),
            _("Jumada l-Ula", "Islamic month lexeme"),
            _("Jumada t-Tania", "Islamic month lexeme"),
            _("Rajab", "Islamic month lexeme"),
            _("Sha`ban", "Islamic month lexeme"),
            _("Ramadan", "Islamic month lexeme"),
            _("Shawwal", "Islamic month lexeme"),
            _("Dhu l-Qa`da", "Islamic month lexeme"),
            _("Dhu l-Hijja", "Islamic month lexeme"),
        )

        self.persian = (
            "",
            # Translators: see
            # http://gramps-project.org/wiki/index.php?title=Translating_Gramps#Translating_dates
            # to learn how to select proper inflection to be used in your localized
            # DateDisplayer code!
            _("Farvardin", "Persian month lexeme"),
            _("Ordibehesht", "Persian month lexeme"),
            _("Khordad", "Persian month lexeme"),
            _("Tir", "Persian month lexeme"),
            _("Mordad", "Persian month lexeme"),
            _("Shahrivar", "Persian month lexeme"),
            _("Mehr", "Persian month lexeme"),
            _("Aban", "Persian month lexeme"),
            _("Azar", "Persian month lexeme"),
            _("Dey", "Persian month lexeme"),
            _("Bahman", "Persian month lexeme"),
            _("Esfand", "Persian month lexeme"),
        )

        self.modifiers = (
            "",
            # Translators: if the modifier is after the date
            # put the space ahead of the word instead of after it
            _("before ", "date modifier"),
            # Translators: if the modifier is after the date
            # put the space ahead of the word instead of after it
            _("after ", "date modifier"),
            # Translators: if the modifier is after the date
            # put the space ahead of the word instead of after it
            _("about ", "date modifier"),
            "",
            "",
            "",
            # Translators: if the modifier is after the date
            # put the space ahead of the word instead of after it
            _("from ", "date modifier"),
            # Translators: if the modifier is after the date
            # put the space ahead of the word instead of after it
            _("to ", "date modifier"),
        )

        self.qualifiers = (
            "",
            _("estimated ", "date quality"),
            _("calculated ", "date quality"),
        )

        # 6753: localized day names. Eventually should sprout into
        # a per-calendar type thing instead.
        self.long_days = (
            "",
            _("Sunday"),
            _("Monday"),
            _("Tuesday"),
            _("Wednesday"),
            _("Thursday"),
            _("Friday"),
            _("Saturday"),
        )

        self.short_days = (
            "",  # Icelandic needs them
            _("Sun"),
            _("Mon"),
            _("Tue"),
            _("Wed"),
            _("Thu"),
            _("Fri"),
            _("Sat"),
        )


# set GRAMPS_RESOURCES then: python3 -m gramps.gen.datehandler._datestrings
if __name__ == "__main__":
    import sys
    from ..utils.grampslocale import GrampsLocale
    from ..const import GRAMPS_LOCALE as glocale
    from ._grampslocale import (
        _deprecated_long_months as old_long,
        _deprecated_short_months as old_short,
        _deprecated_short_days as old_short_days,  # Icelandic needs them
        _deprecated_long_days as old_days,
    )
    from ._datedisplay import DateDisplay
    import gettext

    lang = glocale.lang
    lang_short = lang[:2]
    available_langs = glocale.get_language_list()  # get the cached list
    if glocale.check_available_translations(lang) is None:
        print(
            "Translation for current language {lang} not available.\n"
            "Available translations: {list}.\n"
            "Does po/{lang_short}*.po exist in gramps source tree?!\n"
            "Please set your LANG / LC_ALL environment to something else...\n".format(
                lang=lang, list=available_langs, lang_short=lang_short
            ),
            file=sys.stderr,
        )
        sys.exit(1)

    print(
        "# Generating snippets for {}*.po\n"
        "# Available languages: {}".format(lang_short, available_langs)
    )
    glocale = GrampsLocale(languages=(lang))  # in __main__
    dd = glocale.date_displayer
    ds = dd._ds
    glocale_EN = GrampsLocale(languages=("en"))  # in __main__
    ds_EN = DateStrings(glocale_EN)

    filename = __file__

    try:
        localized_months = dd.__class__.long_months
    except AttributeError:
        localized_months = old_long

    def print_po_snippet(en_loc_old_lists, context):
        for m, localized, old in zip(*en_loc_old_lists):
            if m == "":
                continue
            if m == localized:
                localized = old
            print("#: {file}:{line}".format(file=filename, line=print_po_snippet.line))
            if context:
                print('msgctxt "{context}"'.format(context=context))
            print(
                'msgid "{en_month}"\n'
                'msgstr "{localized_month}"\n'.format(
                    en_month=m, localized_month=localized
                )
            )
            print_po_snippet.line += 1

    print_po_snippet.line = 10000

    try:
        localized_months = dd.__class__.long_months
    except AttributeError:
        localized_months = old_long
    print_po_snippet(
        (ds_EN.long_months, localized_months, old_long),
        "localized lexeme inflections",
    )

    try:
        localized_months = dd.__class__.short_months
    except AttributeError:
        localized_months = old_short
    print_po_snippet(
        (ds_EN.short_months, localized_months, old_short),
        "localized lexeme inflections - short month form",
    )

    print_po_snippet((ds_EN.long_days, old_days, old_days), "")

    print_po_snippet((ds_EN.short_days, old_short_days, old_short_days), "")

    try:
        loc = dd.__class__.hebrew
        print_po_snippet((ds_EN.hebrew, loc, loc), "Hebrew month lexeme")
    except AttributeError:
        pass

    try:
        loc = dd.__class__.french
        print_po_snippet((ds_EN.french, loc, loc), "French month lexeme")
    except AttributeError:
        pass

    try:
        loc = dd.__class__.islamic
        print_po_snippet((ds_EN.islamic, loc, loc), "Islamic month lexeme")
    except AttributeError:
        pass

    try:
        loc = dd.__class__.persian
        print_po_snippet((ds_EN.persian, loc, loc), "Persian month lexeme")
    except AttributeError:
        pass

    try:
        loc = dd.__class__._mod_str
        print_po_snippet((ds_EN.modifiers, loc, loc), "date modifier")
    except AttributeError:
        pass

    try:
        loc = dd.__class__._qual_str
        print_po_snippet((ds_EN.qualifiers, loc, loc), "date quality")
    except AttributeError:
        pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import webbrowser
import sys
from urllib.parse import quote

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.const import URL_MANUAL_PAGE, URL_WIKISTRING
from gramps.gen.constfunc import is_quartz, mac
from gramps.gen.config import config
from .utils import open_file_with_default_application as run_file

# list of manuals on wiki, map locale code to wiki extension, add language codes
# completely, or first part, so pt_BR if Brazilian portugeze wiki manual, and
# nl for Dutch (nl_BE, nl_NL language code)
MANUALS = {
    "de": "/de",
    "fi": "/fi",
    "fr": "/fr",
    "he": "/he",
    "mk": "/mk",
    "nl": "/nl",
    "ru": "/ru",
    "sk": "/sk",
    "sq": "/sq",
}

# first, determine language code, so nl_BE --> wiki /nl
lang = glocale.language[0]
if lang in MANUALS:
    EXTENSION = MANUALS[lang]
else:
    EXTENSION = ""


def display_help(webpage="", section=""):
    """
    Display the specified webpage and section from the Gramps wiki.
    """
    if not webpage:
        link = URL_WIKISTRING + URL_MANUAL_PAGE + EXTENSION
    else:
        link = URL_WIKISTRING + quote(webpage) + EXTENSION
        if section:
            link += "#" + quote(section.replace(" ", "_")).replace("%", ".")
    display_url(link)


def display_url(link, uistate=None):
    """
    Open the specified URL in a browser.
    """
    webbrowser.open_new_tab(link)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Debug/Populate sources and citations"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# gnome/gtk
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import COLON, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gui.utils import ProgressMeter
from gramps.gui.plug import tool
from gramps.gui.dialog import OkDialog
from gramps.gui.managedwindow import ManagedWindow
from gramps.gen.lib import Citation, Source
from gramps.gen.db import DbTxn


class PopulateSources(tool.Tool, ManagedWindow):
    """
    Tool that populates the database with sources and citations.
    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.label = "Populate sources and citations tool"
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.set_window(Gtk.Window(), Gtk.Label(), "")
        tool.Tool.__init__(self, dbstate, options_class, name)

        dialog = self.display()
        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.ACCEPT:
            self.on_ok_clicked()
            OkDialog(
                "Data generated",
                "The requested sources and citations were generated",
                parent=uistate.window,
            )

        self.close()

    def display(self):
        """
        Constructs the GUI, consisting of a message, and fields to enter the
        required number of sources and citations
        """

        # retrieve options
        num_sources = self.options.handler.options_dict["sources"]
        num_citations = self.options.handler.options_dict["citations"]

        # GUI setup:
        dialog = Gtk.Dialog(
            title="Populate sources and citations tool",
            transient_for=self.uistate.window,
            modal=True,
            destroy_with_parent=True,
        )
        dialog.add_buttons(
            _("_Cancel"), Gtk.ResponseType.REJECT, _("_OK"), Gtk.ResponseType.ACCEPT
        )
        label = Gtk.Label(
            label="Enter a valid number of sources and citations."
            " This will create the requested number of sources,"
            " and for each source, will create the requested"
            " number of citations."
        )
        label.set_line_wrap(True)

        hbox1 = Gtk.Box()
        label_sources = Gtk.Label(label="Number of sources" + COLON)
        self.sources_entry = Gtk.Entry()
        self.sources_entry.set_text("%d" % num_sources)
        hbox1.pack_start(label_sources, False, True, 0)
        hbox1.pack_start(self.sources_entry, True, True, 0)

        hbox2 = Gtk.Box()
        label_citations = Gtk.Label(label="Number of citations" + COLON)
        self.citations_entry = Gtk.Entry()
        self.citations_entry.set_text("%d" % num_citations)
        hbox2.pack_start(label_citations, False, True, 0)
        hbox2.pack_start(self.citations_entry, True, True, 0)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(label, True, True, 0)
        vbox.pack_start(hbox1, False, True, 0)
        vbox.pack_start(hbox2, False, True, 0)

        dialog.vbox.set_spacing(10)
        dialog.vbox.pack_start(vbox, True, True, 0)
        dialog.show_all()
        return dialog

    def on_ok_clicked(self):
        """
        Method that is run when you click the OK button. The numbers of sources
        and citations are retrieved from the entry box and used to govern the
        amount of data generated
        """

        num_sources_text = self.sources_entry.get_text()
        try:
            num_sources = int(num_sources_text)
        except:
            return
        num_citations_text = self.citations_entry.get_text()
        num_citations = int(num_citations_text)

        self.progress = ProgressMeter("Generating data", "", parent=self.uistate.window)
        self.progress.set_pass("Generating data", num_sources * num_citations)
        LOG.debug("sources %04d citations %04d" % (num_sources, num_citations))

        source = Source()
        citation = Citation()

        self.db.disable_signals()
        with DbTxn("Populate sources and citations", self.db) as trans:
            for i in range(num_sources):
                source.gramps_id = None
                source.handle = None
                source.title = "Source %04d" % (i + 1)
                source_handle = self.db.add_source(source, trans)

                for j in range(num_citations):
                    citation.gramps_id = None
                    citation.handle = None
                    citation.source_handle = source_handle
                    citation.page = "Page %04d" % (j + 1)
                    self.db.add_citation(citation, trans)
                    self.progress.step()
            LOG.debug("sources and citations added")
        self.db.enable_signals()
        self.db.request_rebuild()
        self.progress.close()

        self.options.handler.options_dict["sources"] = num_sources
        self.options.handler.options_dict["citations"] = num_citations
        # Save options
        self.options.handler.save_options()


class PopulateSourcesOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "sources": 2,
            "citations": 2,
        }
        self.options_help = {
            "sources": ("=num", "Number of sources to generate", "Integer number"),
            "citations": (
                "=num",
                "Number of citations to generate for each source",
                "Integer number",
            ),
        }
